/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Simulator for Sample size calcualtion in animal 
 *                    carcinogenicity studies  built by Hojin Moon, Ph.D. in the 
 *                    department of Biostatistics at M.D. Anderson Cancer Center
 *
 *                    This program conducts Monte Carlo simmulations for various
 *                    animal carcinogenicity experiments to test
 *					  null hypothesis s.t. there is no dose effect 
 *					  among dose groups and alternatives s.t. there is a positive 
 *					  trend among dose groups.
 *                    The Peto test is used for power
 *					  calculations in animal carcinogenicity studies for occult 
 * 					  tumors using Monte Carlo simulation by varying sample sizes. 
 *
 * File Name::        drive.cpp
 * Version::          1.0
 * Type::             source
 * Date Completed::   03/20/2002
 *
 * Environment::
 *     Language::     ANSI C
 *     OS::           Windows 2000
 *      
 * Module Information::
 *     Owner::        M.D. Anderson Cancer Center
 *     Department::   Biostatistics
 *     Researcher::   Hojin Moon, Ph.D.
 *     Application::  ACSSKernel 
 *
 *============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <time.h>
#include <sys/timeb.h>
#include <io.h>        // for  _access
#include <BiostatGeneral.h>
#include "drive.h"
#include "driveP.h"

using namespace MDACC_Biostat;

#define MAX_PATHLEN 260
#define MAX(a,b) ((a) >= (b) ? (a) : (b))

// Local prototype:
int GetUniqueFileName( char const * szDir,  char const * szPrefix, char const *  szExtension, 
                       char * szPathPlusName );
 
/*
 * Global variables
 */

FILE *outfile;

/*
 * Public Routines
 */
static int CompareDbl (const void *pv1, const void *pv2);
static yes_no find_pivot(double *left,double *right,double *pivot_ptr);

/*============================================================================
 *
 * Name:: ACSSFunc
 *
 * Description::  Computes statistical power for detecting a dose-related trend 
 *                in the occurrence of a tumor of interest in the presence of 
 *                competing risks
 *      
 *
 * Inputs::
 *     data -  XML Input struct, containing the input data
 *
 * Outputs::
 *     (None)
 *
 * Return Value:: 0 if successful, otherwise -1
 *     
 *
 *============================================================================*/
 
 
int ACSSFunc (xmlInputDataStruct data)
{
    int nReturn = -1;
	register int i;
	int numGroups;
	int nit;
	int fix_intvl; 
	int interval;
	int test_sides;
    double z_alpha;
	double test_level;

	gendataInputStruct  genInput;
	petotestInputStruct petoInput;
	averages_t          averageInput;
	sampleData_t*       sampData;  
	 

	// create a tempfile for output. Note: modify if you need to preserve the output file
	if((outfile = tmpfile()) == NULL)
		throw "Could not create outputfile";

	// Perform input validation, output to stdio, populate the input structs
	InitialProcess(data, &genInput, &petoInput, &averageInput );

    // initialize local variables 
    numGroups = genInput.dose_group;
	nit       = data.numberDataSets;
    fix_intvl = genInput.fix_intvl;
 
	// Allocate memory for two arrays of structs of size #groups
	sampData   = (sampleData_t *)malloc(numGroups *sizeof(sampleData_t));
 
 
     // allocate memory for the struct array elements
     AllocateDataStructs( &sampData,  numGroups,  petoInput.maxintvl);

	 // calculate z_alpha from the InverseCDF of the standard Normal distribution	
	 test_level = petoInput.test_level;
	 test_sides = petoInput.test_side;

	 Normal mynorm( 0.0, 1.0 );
	 z_alpha = fabs( mynorm.CDFInverse( test_level/(double) test_sides) );

	// start the process
     
    for( i = 0; i < nit; i++)
	{
	    if (gendata( genInput, sampData, &averageInput, &interval) != -1)
            nReturn = petotest(petoInput, interval, z_alpha);
		 
	}

	/* Get statistics */
	averageInput.nit = nit;
	GetAverages( averageInput );

	/* Calculate and output power */
	CalculatePower();
 
	// free local pointers
    DeallocateDataStructs( sampData, numGroups );

	free ( sampData );
	 

	// free globals
    FinalProcess( numGroups, genInput, petoInput, averageInput );

	// close outfile
	_rmtmp();

return nReturn;  
}



/*
 * Private Routines
 */

/*============================================================================
 *
 * Name:: petotest
 *
 * Description:: Implementation of the  Peto test to compute power to detect 
 *               dose-related trend in the incidence of tumor of interest.
 *      
 *
 * Inputs::
 *     petoInput -  petotest Input struct, containing the input data
 *     intvlft   -  integer containing the fatal interval
 *     z_alpha   -  double containg the z-value at the probability of a Type I error rate alpha
 *
 *
 * Return Value:: EXIT_SUCCESS if successful, otherwise EXIT_FAILURE
 *     
 *
 *============================================================================*/
int petotest( const petotestInputStruct_t petoInput,
			  int intvlft, double z_alpha )
{	 
    int j, k, den_flg, fg1, fg2, fg3, mark1, mark2, interval;
	int psdflg;
	int group; 
    int maxintvl;
	int side_test;
	int* animal;
	double poly_div;
	double test_level;
    double * m_2;
	double * tmp_D;
	double** tmp_V;
	double *kappa;
	double *kappa_ft;
	int maxintvl_length = 0; 
	
	/* initialize local variables */
    poly_div   = 0.0;
    mark1      = 0; 
	mark2      = 0;
	fg1        = 0;
    fg2        = 0;
	group      = petoInput.numGroups;
    maxintvl   = petoInput.maxintvl;
	side_test  = petoInput.test_side;
	test_level = petoInput.test_level;
	animal     = petoInput.animal;


	/* for incidental part */
	
    m_2   = allocate_real_vector(0,group-1);
	tmp_D = allocate_real_vector(0,group-1);
	tmp_V = memory_alloc_mat(0,group-1,0,group-1);

	/* Find max(intvl_inc, intvlft ) */

	maxintvl_length = MAX( intvl_inc, intvlft );
	kappa = allocate_real_vector( 0, maxintvl_length - 1 );
	kappa_ft = allocate_real_vector( 0, maxintvl_length - 1);


	interval = intvl_inc;
	den_flg  = cal_expect(0, interval, group);
	if(den_flg != 0)
	{ 	
		make_D_vec( 0, interval, group );
		fg1 = cal_V_matrix( 0, interval, group, kappa, kappa_ft );
	}
	if(den_flg == 0 || fg1 == 0)
	{	
		for(j = 0; j < group; j++)
		{
			D[j] = 0.0;
			for(k = 0; k < group; k++)	
				V_a[j][k] = 0.0;
		}	
	}
	if(a1_flag != 0 && den_flg != 0 && fg1 != 0)
	{
		psdflg = calc_stat( 0, group, m_2, tmp_D, tmp_V, z_alpha );
		if(psdflg != 0)
		{  
			  nit_inc += 1.0; mark1 = 1;
		}
	}

	/* for fatal */
	interval = intvlft;

	den_flg  = cal_expect(1, interval, group); 
	if(den_flg != 0)
	{	
		make_D_vec(1, interval, group);
		fg2 = cal_V_matrix(1, interval, group, kappa, kappa_ft);
	}
	if(den_flg == 0 || fg2 == 0)
	{
		for(j = 0; j < group; j++)
		{
			D_ft[j] = 0.0;
            for(k = 0; k < group; k++)     
				V_b[j][k] = 0.0;
         }
    }
	if(d_flag != 0 && den_flg != 0 && fg2 != 0)
	{	
		    psdflg = calc_stat( 1, group, m_2, tmp_D, tmp_V, z_alpha );
		if(psdflg != 0)
		{  
			 nit_ft += 1.0; mark2 = 1;
		}
	}

	/* interval is not meaningful in here */
	interval = intvl_inc; 
	make_D_vec(2, interval, group); /* for fatal + incidental */
	fg3      = cal_V_matrix( 2, interval, group, kappa, kappa_ft );
	if(mark1 || mark2)
	{
	    psdflg = calc_stat( 2, group, m_2, tmp_D, tmp_V, z_alpha );
		if(psdflg != 0)
		{  
			  nit_comb += 1.0;	
		}
	}
 
	/* deallocate local arrays */
	free_real_vector(m_2,0);
	free_real_vector(tmp_D,0);
	free_real_matrix(tmp_V,0,group-1,0);
	free_real_vector(kappa,0);
	free_real_vector(kappa_ft,0);

	return EXIT_SUCCESS;
}

 
/*============================================================================
 *
 * Name:: calc_stat
 *
 * Description:: Calculate the Peto test statistics 
 *      
 *
 * Inputs::
 *     cnt         -   tumor indicator
 *     group       -   number of groups
 *	   m_2         -   intermediate array of double 
 *     tmp_D       -   intermediate array of double 
 *     tmp_V       -   intermediate two dimensional array of double 
 *     z_alpha     -   double containg the z-value at the probability of a Type 
 *                     I error rate alpha  
 *
 * Return Value::
 *     0 or 1  
 *
 *============================================================================*/

int calc_stat(int cnt, 
			  int group, 
			  double *m_2, 
			  double *tmp_D,
	          double **tmp_V,
			  double z_alpha)
{
    register int i,j,k;
	double num, den, X_r;
	int it;
    
	if(cnt == 0)
	{ 
		for(i = 0; i < group; i++)
		{
			for(j = 0; j < group; j++)
				tmp_V[i][j] = V_a[i][j];
			tmp_D[i] = D[i];
		}
	}
	else if(cnt == 1)
	{
		for(i = 0; i < group; i++)
		{
			for(j = 0; j < group; j++)
				tmp_V[i][j] = V_b[i][j];
			tmp_D[i] = D_ft[i];
		}
	}
	else
	{
		for(i = 0; i < group; i++)
		{
			for(j = 0; j < group; j++)
				tmp_V[i][j] = V_tot[i][j];
			tmp_D[i] = D_tot[i];
		}
	}

	/* Positive Trend Test */
	num = 0.0;
	for(i = 0; i < group; i++)
		num += dose[i]*tmp_D[i];
	for(i = 0; i < group; i++)
	{	
		m_2[i] = 0.0;
		for(k = 0; k < group; k++)
		{	
			m_2[i] += tmp_V[i][k]*dose[k];
		}
	}
	den = 0.0;
	for(k = 0; k < group; k++)	den += dose[k]*m_2[k];
	if(den == 0)   return(0);

	X_r = num/sqrt(den);
 
	it = 0;
    if ( X_r > z_alpha ) 
		it = 1;
	 if( it == 1 )
	 {
		if ( cnt == 0 ) pwr_trend_inc += 1.0;
		else if ( cnt == 1 ) pwr_trend_ft += 1.0;
		else if ( cnt == 2 ) pwr_trend_comb += 1.0;
		else throw ("Error in calc_stat() cnt entry");
	 }
	
	return(1);
}


/*============================================================================
 *
 * Name:: cal_V_matrix
 *
 * Description:: calculate gxg matrix covariance matrix in the test
 *      
 *
 * Inputs::
 *     flag        -   tumor indicator
 *     intvl       -   number of interval
 *     group       -   number of groups
 *     kappa       -   intermediate vector for variance calculation  
 *     kappa_ft    -   intermediate vector for variance calculation    
 *
 * Outputs::
 *     (none)
 *
 * Return Value::
 *     0 or 1  
 *============================================================================*/

int cal_V_matrix(int flag, int intvl, int group, double *kappa, double *kappa_ft)
{
 	register int i;
	register int r;
	register int j; 
	int delta;
	double sum;
	
	if(flag == 0)
	{
		for(j = 0; j < intvl; j++)
		{
			if((n_g_sum[j] - 1) == 0)
			{
				 
				return(0);
			}
			kappa[j] = y_g_sum[j]*(n_g_sum[j] - y_g_sum[j]) /
				(double) (n_g_sum[j] - 1.0);
		}
		for(r = 0; r < group; r++)
		{
			for(i = 0; i < group; i++)
			{
				sum = 0.0;
				if(r == i)	delta = 1;
				else delta = 0;
				for(j = 0; j < intvl; j++)
					sum += kappa[j]*K[r][j]*(delta-K[i][j]);
				V_a[r][i] = sum;
			}
		}
		
		return(1);
	}/* end of if flag == 0 */
	else if(flag == 1)
	{
		for(j = 0; j < intvl; j++)
		{
			if((n_g_sum_ft[j] - 1) == 0)
			{
				
				return(0);
			}
			kappa_ft[j] = y_g_sum_ft[j]*(n_g_sum_ft[j] -
				 y_g_sum_ft[j])/(double) (n_g_sum_ft[j] - 1.0);
		}
		for(r = 0; r < group; r++)
			for(i = 0; i < group; i++)
			{
				sum = 0.0;
				if(r == i)	delta = 1;
				else delta = 0;
				for(j = 0; j < intvl; j++)
					sum+=kappa_ft[j]*K_ft[r][j]*
						(delta-K_ft[i][j]);
				V_b[r][i] = sum;
			}
	
		return(1);
	}
	else
	{
		for(i = 0; i < group; i++)
			for(j = 0; j < group; j++)
			    V_tot[i][j] = V_a[i][j] + V_b[i][j]; 
		
		return(1);
	}
}

/*============================================================================
 *
 * Name:: make_D_vec
 *
 * Description:: make gx1 vector D (difference vector between observed
 *               and expected number of tumors
 *      
 *
 * Inputs::
 *     flag        -   tumor indicator
 *     intvl       -   number of intervals
 *     group       -   number of groups
 * 
 *
 *
 * Return Value::
 *     (none)
 *
 *============================================================================*/


void make_D_vec(int flag, int intvl, int group)
{
 
	register int i;
	register int j;
	int sum1;
	double sum2;	
	
	if(flag == 0){
		for(i = 0; i < group; i++){
			sum1 = 0; sum2 = 0.0;
			for(j = 0; j < intvl; j++){
				sum1 += no_w_it[i][j];
				sum2 += E[i][j];
			}
			D[i] = sum1 - sum2;
		}
	}/* end of if flag == 0 */
	else if(flag == 1){
		for(i = 0; i < group; i++){
			sum1 = 0; sum2 = 0.0;
			for(j = 0; j < intvl; j++){
				sum1 += no_w_t[i][j];
				sum2 += E_ft[i][j];
			}
			D_ft[i] = sum1 - sum2;
		}
	}
	else{
		for(i = 0; i < group; i++)
			D_tot[i] = D[i] + D_ft[i];
	}
}


/*============================================================================
 *
 * Name:: cal_expect
 *
 * Description:: A part of a routine for computing expected number of
 *               tumors according to flag ie, flag 0 stands 
 *               for incidental and flag 1 for fatal tumor
 *
 * Inputs::
 *     flag        -   tumor indicator
 *     intvl       -   number of intervals
 *     group       -   number of groups
 * 
 * Outputs::
 *     (none) 
 *
 * Return Value::
 *     0 or 1 
 *
 *============================================================================*/

int cal_expect(int flag, int intvl, int group)
{
 	register int i;
	register int j;
	int sum1, sum2;
	
	if(flag == 0){
	    for(j = 0; j < intvl; j++){
		sum1 = 0; sum2 = 0;
		for(i = 0; i < group; i++){
			sum1 += no_w_it[i][j];	
			sum2 += no_dead_it[i][j];	
		}
		y_g_sum[j] = sum1;
		n_g_sum[j] = sum2;
	    }
	    for(i = 0; i < group; i++)
		for(j = 0; j < intvl; j++){
			if(n_g_sum[j] == 0)	return(0);
			K[i][j] = no_dead_it[i][j]/(double) n_g_sum[j];
		}
	    for(i = 0; i < group; i++)
		for(j = 0; j < intvl; j++)
			E[i][j] = y_g_sum[j]*K[i][j];
	
	    return(1);
	}/* end of if flag == 0 */

	else{ /* for the case of flag == 1 */
	    for(j = 0; j < intvl; j++){
		sum1 = 0; sum2 = 0;
		for(i = 0; i < group; i++){
			sum1 += no_w_t[i][j];	
			sum2 += no_surv[i][j];	
		}
		y_g_sum_ft[j] = sum1;
		n_g_sum_ft[j] = sum2;
	    }
	    for(i = 0; i < group; i++)
		for(j = 0; j < intvl; j++){
			if(n_g_sum_ft[j] == 0)	return(0);
			K_ft[i][j] = no_surv[i][j]/(double) n_g_sum_ft[j];
		}
	    for(i = 0; i < group; i++)
		for(j = 0; j < intvl; j++)
			E_ft[i][j] = y_g_sum_ft[j]*K_ft[i][j];
	
 	    return(1);
	}/* end of else flag == 1 */
}

/*============================================================================
 *
 * Name:: init_incidental
 *
 * Description:: Get input data for incidental tumors
 *      
 *
 * Inputs::
 *     group      -   number of groups
 *     sampData   -   pointer to a sampleData struct 
 *
 * Return Value::
 *     (none)
 *
 *============================================================================*/
void init_incidental(int group, sampleData* sampData )
{
        int d, a1, a2, b1, b2;
        int i,j, sum_inc;

        sum_inc = 0;
        for(i = 0; i < group; i++)
		{
			for(j = 0; j < intvl_inc; j++)
			{
                
				d  = sampData[i].numberDead[j+1];
				a1 = sampData[i].numDeadWithTumor[j+1];
				b1 = sampData[i].numDeadWithoutTumor[j+1];
				a2 = sampData[i].numSacrificesWithTumor[j+1];
				b2 = sampData[i].numSacrificesWithoutTumor[j+1];

                no_w_it[i][j] = a1+a2;
                no_dead_it[i][j] = a1+b1+a2+b2;
                sum_inc += a1 + a2;
		 
			}
			 
        }
        if(sum_inc == 0) 
			a1_flag = 0;
        else    
			a1_flag = 1;
}

/*============================================================================
 *
 * Name:: init_fatal
 *
 * Description:: Get input data for fatal tumors
 * 
 * Inputs::
 *     group      -   number of groups
 *     maxintvl   -   maximum number of intervals
 *     sampData   -   pointer to a sampleData struct 
 *
 * Return Value::
 *     (none)
 *
 *============================================================================*/


void init_fatal(int group, int maxintvl, sampleData* sampftData )
{
	register int i;
	register int j;
	register int k;
    int d, a1, a2, b1, b2, *tot;
	int intvl = 0;
	int sum_d;

 	tot = allocate_integer_vector(0,maxintvl-1);
	
	sum_d = 0;

	for(i = 0; i < group; i++)
	{
	    intvl = sampftData[i].fixedinterval;
		
	    for(j = 0; j < intvl; j++)
		{    
			
			d  = sampftData[i].numberDead[j+1];
			a1 = sampftData[i].numDeadWithTumor[j+1];
			b1 = sampftData[i].numDeadWithoutTumor[j+1];
			a2 = sampftData[i].numSacrificesWithTumor[j+1];
			b2 = sampftData[i].numSacrificesWithoutTumor[j+1];

			tot[j]=d+a1+b1+a2+b2;
			no_w_t[i][j] = d;
			sum_d += d;
	    }
	     
	    for(j=0; j<intvl; j++)
		{
	        no_surv[i][j]=0;  
			for(k=j;k<intvl;k++)  
				no_surv[i][j] += tot[k];
	    }
	}

	if(sum_d == 0)	
		d_flag = 0;
	else	
		d_flag = 1; 

	free_integer_vector(tot,0);

	return;
}


/*============================================================================
 *
 * Name:: gendata
 *
 * Description:: Generation of simulation data according to input parameters
 *      
 *
 * Inputs::
 *      data           -   struct containing the input data extracted from the XML string
 *      sampdata       -   pointer to the array of sampleData_t structs
 *      averageInput   -   pointer to averageInput struct holding the data for generating
 *                         the averages printed at the end of the run
 *      interval       -   number of intervals ( for init_fatal )
 * 
 *
 * Return Value:: EXIT_SUCCESS if successful, otherwise EXIT_FAILURE
 *     
 *
 *============================================================================*/

int gendata( const gendataInputStruct genInput, 
			 sampleData_t * sampdata, 
			 averages_t * averageInput, 
			 int * interval)
 {
	register int i;
	register int j;
	register int k;
	int m, jj, sect, intvl_no, isac, idie;
	int *index;
	int **no_sac_int;
	int NTP, dose_group, rn, N_tot;
	double aa, bb, u, eta, tt_cnt, t1_cnt, t3_cnt, phi2, phi, pt3, t2;
	double pt1, delta2;
	double *ed; /* represents e^d */
	double *comp; /* CRSR along with dose groups */
	double *tpt1; /* holds Tumor Onset Pr. of 4 dose groups */
	double *intvl_arry, *reduced_arry, *dose_level;
    int intvl_flg;
	int fix_intvl;
	int maxintvl;
	int maxNumPerGroup;
	double * temp;
	double * A;
    int * N;
	int * sactime;
	double * avg_leth; 
	double * avg_t1;
	double * avg_t3; 
	double * divisor;
	int** d_avg; 
	int** a1_avg; 
	int** b1_avg;
	int** a2_avg;
	int** b2_avg;
     
    if ( averageInput != NULL)
	{
		N        = averageInput->N;
		sactime  = averageInput->sactime;
		avg_leth = averageInput->avg_leth;
		avg_t1   = averageInput->avg_t1;
		avg_t3   = averageInput->avg_t3;
		divisor  = averageInput->divisor;
		d_avg    = averageInput->d_avg;
		a1_avg   = averageInput->a1_avg;
		b1_avg   = averageInput->b1_avg;
		a2_avg   = averageInput->a2_avg;
		b2_avg   = averageInput->b2_avg;
	}
	else
	{
		return EXIT_FAILURE;
	}

	/* Extract values form genInput */	
	
	dose_group = genInput.dose_group;  
	intvl_flg  = genInput.intvl_flg;
	fix_intvl  = genInput.fix_intvl;
	maxintvl   = genInput.maxintvl;
	rn         = genInput.rn;
	N_tot      = genInput.N_tot;
	NTP        = genInput.NTP; /* number of sacrifices including a terminal sacrifice */
	pt1        = genInput.pt1; /* background tumor onset probability */
	delta2     = genInput.delta2; /* Tumor onset distribution */
	phi2       = genInput.phi2; /* tumor lethality */
    dose_level = genInput.dose_level;
	ed         = genInput.ed;
	comp       = genInput.comp;
	tpt1       = genInput.tpt1;
	no_sac_int = genInput.no_sac_int;
 
    /* initialize N and find maxNumPerGroup */
	maxNumPerGroup = 0;
	for ( i = 0; i < dose_group; i++)
	{
		N[i]  = genInput.N[i];
		if ( N[i] > maxNumPerGroup )
			maxNumPerGroup = N[i];
	}
 
 	temp = allocate_real_vector(0, maxNumPerGroup );
	A = allocate_real_vector(0,  maxNumPerGroup);
    index = allocate_integer_vector(0,maxNumPerGroup);

	for(jj = 0; jj < dose_group; jj++)
	{
 
		gen_animal_sort(index, temp, A, N[jj]);
		tt_cnt = 0.0; 
		t1_cnt = 0.0; 
		t3_cnt = 0.0;

		phi = 1.0;
		pt3 = comp[jj];
    
		if(pt3 < f_gam) 
			phi = log(pt3)/log(f_gam);
		eta = ed[jj];
		sect = 0;
		
		for(k=0;k<jj;k++)  
			sect += N[k];
 
		for(i = 0; i < N[jj]; i++)
		{
			if (NTP > 1)
			{
		   /* each animal has its own sac. time */
				for(j=0;j<NTP;j++)
					if(index[i] <= no_sac_int[jj][j])
					{
						animals[sect+i].sac = sactime[j];
						j=NTP+1;
					}
			
			/* convert to days */
				 animals[sect+i].sac *= 7.0;
			}
			else if (NTP ==1)
			{
				animals[sect+i].sac = rn * 7.0;
			}
			else
			{   
				 throw "Error in input NTP!\n";
			}
			/* Tumor onset time (T1) */
			do
			{     
				u = getUniform();
			}
			while(u <= 1.0e-99);
			
			animals[sect+i].t1 = rn*pow(-log(u)/(eta*delta1),1.0/delta2);
			if(animals[sect+i].t1 < ((double) rn))
				++t1_cnt;

			animals[sect+i].t1 *= 7.0;

			/* Time to death from competing risks (T3) */
			do
			{     
				u = getUniform();
			}
			while(u <= 1.0e-99);
			
			if( u <= pow(f_gam, phi) )
			{
				animals[sect+i].t3 = rn + 1.0;
			}
			else
			{   
			  aa = 0.0; 
			  bb = (double) rn;
			  animals[sect+i].t3 = zbrent ( u, aa, bb, phi, rn );  
			}
			

			if(animals[sect+i].t3 < ((double) rn))
				++t3_cnt;

			animals[sect+i].t3 *= 7.0;

			/* Time to death from tumor after onset (T2) */
			do
			{   
				u = getUniform();
			}
			while(u <= 1.0e-99);
			
			if(u <= pow(f_gam, phi2))       
				t2 = rn+1.0;
			else
			{   
				aa = 0.0; 
				bb = (double) rn;
				t2 = zbrent ( u, aa, bb, phi2, rn );  
			}
			t2 *= 7.0;

		 
			animals[sect+i].tt = animals[sect+i].t1 + t2;
			if(animals[sect+i].tt < rn*7.0) 
				++tt_cnt;

			/* Time of natural death */
			if(animals[sect+i].t3 <= animals[sect+i].tt)
				animals[sect+i].die=animals[sect+i].t3;
			else	
				animals[sect+i].die=animals[sect+i].tt;
				
			if((animals[sect+i].sac>animals[sect+i].die) &&
			   (animals[sect+i].t1<=animals[sect+i].die) &&
			   (animals[sect+i].tt <= animals[sect+i].t3))
			{
				animals[sect+i].d_intvl_time = animals[sect+i].tt;
			}
			else
			{
				animals[sect+i].d_intvl_time = 0.0;
			}
		} /* for i = 1 .. N */

		if(t1_cnt != 0.0)
		{
			avg_leth[jj] += tt_cnt/t1_cnt;
			++divisor[jj];
		}
		 avg_t1[jj] += t1_cnt/ N[jj];
		 avg_t3[jj] += t3_cnt/ N[jj];
 
	} /* for jj = 1 .. g */

 
	/* INCIDENTAL part consists of NTP intervals */
	if(NTP > 1)
	{
		intvl_no = NTP;
		intvl_arry = allocate_real_vector(1,intvl_no);
		assgn_arry(intvl_arry, sactime, NTP);
		
	}
	else if(NTP ==1)
	{
		intvl_no = 4;
		intvl_arry = allocate_real_vector(1,intvl_no);
		assgn_arry(intvl_arry, sactime, intvl_no);
		
	}
	else
	{
		throw "Error in input NTP!\n";
    }
	reduced_arry = allocate_real_vector(1, fix_intvl);
    try
	{
		 assign_fix_intvl(intvl_arry,reduced_arry,intvl_no,fix_intvl);
	}
	catch (const char * s)
	{
        if (strcmp(s,"fatal error in assign_fix_intvl") == 0)
			throw " error in assigning fixed intervals!\n";
	}
	/* now we have the array which holds the NTP intervals, 
	   so we only need reduced array from now on.
	   Therefore, get rid of the interval array now. 
	*/

	free_real_vector(intvl_arry, 1);
	
	/* Construct incidental data set fro each group */
 
	for(jj = 0; jj < dose_group; jj++)
	{
		
		for(m = 1; m <= fix_intvl; m++)
		{
            sampdata[jj].numberDead[m]                = 0; 
			sampdata[jj].numDeadWithTumor[m]          = 0; 
			sampdata[jj].numDeadWithoutTumor[m]       = 0;
            sampdata[jj].numSacrificesWithTumor[m]    = 0; 
			sampdata[jj].numSacrificesWithoutTumor[m] = 0;
			sampdata[jj].intervalEndpoints[m]         = reduced_arry[m];
		}
		 
		sampdata[jj].dose_level = dose_level[jj];

	    sect = 0;
		for( k=0; k<jj; k++ )  
			sect += N[k];

		for(i = 0; i < N[jj]; i ++)
		{
			for(j = 1; j <= fix_intvl; j++)
			{
				if(animals[sect+i].sac <= sampdata[jj].intervalEndpoints[j])
				{
					isac = j; 
					j    = fix_intvl+1;
				}
				else	
					isac = fix_intvl;
			}
			for(j = 1; j <= fix_intvl; j++)
			{
				if(animals[sect+i].die <= sampdata[jj].intervalEndpoints[j])
				{
					idie = j;
					j    = fix_intvl+1;
				}
				else	
					idie = fix_intvl;
			}

			/* Construction of data set */
			if(animals[sect+i].sac <= animals[sect+i].die)
			{
				/* sacrificed */
				if(animals[sect+i].t1 <= animals[sect+i].sac)
				{
					/* sac. w/ tumor */
					++(sampdata[jj].numSacrificesWithTumor[isac]); 
				 
				}
				else
				{ 
					/* sac. w/o tumor */
					++sampdata[jj].numSacrificesWithoutTumor[isac];
					 
				}
			}
			else
			{
				/* natural death */
				if(animals[sect+i].t1 <= animals[sect+i].die)
				{
					/* natural death w/ tumor */
					if(animals[sect+i].tt <= animals[sect+i].t3)
					{
						/* death w/ fatal */
						++(sampdata[jj].numberDead[idie]); 
						 
					}
					else
					{
						/* death w/ incidental */
						++(sampdata[jj].numDeadWithTumor[idie]);
					 
					} /* end else statement */
				}
				else
				{
					/* natural death w/o tumor */
					++(sampdata[jj].numDeadWithoutTumor[idie]);
					 
				}
			}
			 
		} /* for i = 0 .. N */

 		/* To compute avg death rates */
		for (k = 0; k < intvl_flg; k++)
		{
			 
			 d_avg[jj][k]  += sampdata[jj].numberDead[k+1];
			 a1_avg[jj][k] += sampdata[jj].numDeadWithTumor[k+1];
			 b1_avg[jj][k] += sampdata[jj].numDeadWithoutTumor[k+1];
			 a2_avg[jj][k] += sampdata[jj].numSacrificesWithTumor[k+1];
			 b2_avg[jj][k] += sampdata[jj].numSacrificesWithoutTumor[k+1];

		}
		/* updated on 3/4/02 */ 
	} /* for jj = 0 .. g */

		// call init_incidental to update globals
		init_incidental(dose_group, sampdata);

		free_real_vector(reduced_arry,1);
		 
		/* generate data for FATAL part w/o NPT intvl */
		intvl_no   = det_no_of_intvl_ft(rn,N_tot);
		intvl_arry = allocate_real_vector(1,intvl_no);
		assgn_arry_ft(intvl_arry, intvl_no, rn, N_tot);
		
		if(intvl_no != 1)
		{	
			quicksort(intvl_arry+1, intvl_arry+intvl_no); 
		    roundoff_arry (intvl_arry, intvl_no);
            fix_intvl = decide_intvl (intvl_arry, intvl_no);
			 
         }
         else    
			 fix_intvl = 1;
		
		reduced_arry = allocate_real_vector(1, fix_intvl);
        try
        { 
		    assign_fix_intvl(intvl_arry,reduced_arry,intvl_no,fix_intvl);
		}
		catch ( const char * s)
        {
			if (strcmp(s,"fatal error in assign_fix_intvl") == 0)
			throw " error in assigning fixed intervals!\n";
		}
		free_real_vector(intvl_arry, 1);
		
	
	    // update interval
		(*interval) = fix_intvl;
		
		/* Construct fatal data set for each group */
	   
		for(jj = 0; jj < dose_group; jj++)
		{
   	        sampdata[jj].fixedinterval = fix_intvl;
		    for(m = 1; m <= fix_intvl; m++)
			{
				sampdata[jj].numberDead[m]                = 0; 
				sampdata[jj].numDeadWithTumor[m]          = 0; 
				sampdata[jj].numDeadWithoutTumor[m]       = 0;
                sampdata[jj].numSacrificesWithTumor[m]    = 0; 
			    sampdata[jj].numSacrificesWithoutTumor[m] = 0;
				sampdata[jj].intervalEndpoints[m]         = reduced_arry[m];
             }
			 
			sampdata[jj].dose_level = dose_level[jj];
			sect = 0;
			for (k=0;k<jj;k++)  
				sect += N[k];

			for (i = 0; i < N[jj]; i ++)
			{
				for (j = 1; j <= fix_intvl; j++)
				{
					if(animals[sect+i].sac <= sampdata[jj].intervalEndpoints[j])
					{
						isac = j; 
						j = fix_intvl+1;
					}
					else   
						isac = fix_intvl;
				}
				for (j = 1; j <= fix_intvl; j++)
				{
					if(animals[sect+i].die <= sampdata[jj].intervalEndpoints[j])
					{
						idie = j; 
						j = fix_intvl+1;
					}
					else    
						idie = fix_intvl;
				}
				if(animals[sect+i].sac<=animals[sect+i].die)
				{
					if(animals[sect+i].t1 <= animals[sect+i].sac)
					{
						++(sampdata[jj].numSacrificesWithTumor[isac]);
						 
					}
					else
					{
						++(sampdata[jj].numSacrificesWithoutTumor[isac]);
						 
					}
				}
				else
				{	
					if(animals[sect+i].t1 <= animals[sect+i].die)
					{
						if(animals[sect+i].tt <= animals[sect+i].t3)
						{
							++(sampdata[jj].numberDead[idie]);
							 
						}
						else
						{	
							++(sampdata[jj].numDeadWithTumor[idie]);
							
						}
					}
					else
					{	
						++(sampdata[jj].numDeadWithoutTumor[idie]);
						 
					}
				}
			 
			}
	}/* end for ( jj ..) */
	
	/* call init_fatal to update globals */
	init_fatal(dose_group, maxintvl, sampdata);

	/* free local pointers */
	free_real_vector(reduced_arry,1);
	free_integer_vector(index,0);
	free_real_vector(temp,0);
	free_real_vector(A,0);
 
	  
    return EXIT_SUCCESS;
}


/*============================================================================
 *
 * Name:: assign_fix_intvl
 *
 * Description:: Assign the interval time from interval array to reduced array
 *      
 *
 * Inputs::
 *      intvl_arry         -  interval time points
 *      reduced_arry       -  unique interval timepoints
 *      m                  -  number of intervals
 *      n                  -  reduced number of intervals
 * 
 * Outputs::
 *     (none)
 *
 * Return Value:: EXIT_SUCCESS if successful, otherwise EXIT_FAILURE
 *      
 *
 *============================================================================*/
 

int assign_fix_intvl(double intvl_arry[], double reduced_arry[], int m, int n)
{
 
	register int i;
	int cnt;
    double old, newValue;

    cnt = 1;
    old = intvl_arry[1];
	reduced_arry[1] = intvl_arry[1];
    
	for(i = 2; i <= m; i++)
	{
		newValue = intvl_arry[i];
        if(newValue != old)
		{	
			++cnt;
			reduced_arry[cnt] = newValue;
		}	
        old = newValue;
    }
	if(cnt > n) 
	{
        throw "fatal error in assign_fix_intvl";
    }
    return EXIT_SUCCESS;
}


/*============================================================================
 *
 * Name:: decide_intvl
 *
 * Description:: Define number of intervals for incidental tumor
 *      
 *
 * Inputs::
 *      intvl_arry         -  interval time points 
 *      m                  -  number of intervals
 * Outputs::
 *     (none)
 *
 * Return Value::          - number of intervals
 *      
 *
 *============================================================================*/

int decide_intvl(double intvl_arry[], int m)
{
    register int i;	
	int cnt;
	double old, newValue;

	cnt = 1;
	old = intvl_arry[1];
	for(i = 2; i <= m; i++)
	{
		newValue = intvl_arry[i];
		if(newValue != old)	++cnt;
		old = newValue;
	}
	
	return(cnt);
}

/*============================================================================
 *
 * Name:: roundoff_arry
 *
 * Description:: Generate ceiling function to decide the interval
 *      
 *
 * Inputs::
 *      intvl_arry         -  interval time points  
 *      m                  -  number of intervals
 *
 * Outputs::  
 *      intvl_arry 
 *
 * Return Value::
 *     (none) 
 *
 *============================================================================*/

void roundoff_arry(double intvl_arry[], int m)
{
	register int i;
	 
    for(i = 1; i < m; i++)
	{
		if((intvl_arry[i] - (int) intvl_arry[i]) != 0) 
			intvl_arry[i] = (int) intvl_arry[i] + 1.0;
	}
}

/*============================================================================
 *
 * Name:: assgn_arry_ft
 *
 * Description:: Assign d_int_time to int_t_arrray (assign actual
 *               death time to an interval array) 
 *      
 *
 * Inputs::
 *      intvl_arry         -  interval time points 
 *      m                  -  number of intervals
 *      rn                 -  duration of study
 *      N_tot              -  total number of animals 
 *
 * Outputs::
 *      intvl_arry
 *
 * Return Value::
 *     (none) 
 *
 *============================================================================*/

void assgn_arry_ft(double intvl_arry[], int m, int rn, int N_tot)
{
	int i, k;

    k = 1;
  	intvl_arry[k] = rn*7.0; ++k;
    for(i = 0; i < N_tot; i++)
	{
		if((animals[i].d_intvl_time != 0.0) && (animals[i].d_intvl_time <= rn*7.0))
		{
                        intvl_arry[k] = animals[i].d_intvl_time;
                        ++k;
        }
        if(k > m+1)  
			printf("fatal error to allocate intvl_t_arry\n");
     }
}


/*============================================================================
 *
 * Name:: assgn_arry
 *
 * Description:: Assign NTP intervals
 *      
 *
 * Inputs::
 *      intvl_arry         -  interval time points  
 *      sactime            -  sacrifice time point
 *      NTP                -  number of intervals 
  
 * Outputs::
 *     intvl_arry
 *
 * Return Value::
 *     (none) 
 *
 *============================================================================*/

void assgn_arry(double intvl_arry[], int sactime[], int NTP)
{ 
    register int k;
	
	for (k = 1; k <= NTP; k++)  
		intvl_arry[k]=sactime[k-1]*7.0;
}



/*============================================================================
 *
 * Name:: det_no_of_intvl_ft
 *
 * Description::  Determine number of intervals for the set of g groups 
 *                (determine number of intervals for fatal tumor)
 *
 * Inputs::
 *       rn            -  study duration 
 *       N_tot         -  total number of animals
 *
 * Return Value::
 *      the number of intervals  
 *
 *============================================================================*/

int det_no_of_intvl_ft(int rn, int N_tot)
{
	int m, i;

	m = 1;
    for(i = 0; i < N_tot; i++)
	{
		if((animals[i].d_intvl_time != 0.0)  && (animals[i].d_intvl_time <= rn*7.0))
                        ++m;
	}
    return(m);
}

 
/*============================================================================
 *
 * Name:: zbrent
 *
 * Description::   Brent's method to find zero of a monotone function 
 *      
 * Note:: Modified version of zbrent function from pages 345-365 of Second Edition
 * "Numerical Recipies in C - The Art of Scientific Computing",  by Press et al.(1988)
 *============================================================================*/

#define NRANSI
#define ITMAX 100
#define EPS 3.0e-6
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
   
double zbrent( double u, double x1, double x2, double phi, int rn)
{

	double func( double t, double u, double phi );
 
	double a = x1;
	double b = x2;
	double c = x2;
	double d = 0;
	double e = 0;
	double min1;
	double min2;
	double fa;
	double fb;
	double fc;
	double p;
	double q;
	double r;
	double s;
	double tol1;
	double xm;
	int iter = 0;
   
	fa = func( a, u, phi );
	fb = func( b, u, phi );
   
	if ( ( fa > 0.0 && fb > 0.0 ) || ( fa < 0.0 && fb < 0.0 ) )
	{
 		return( rn + 1.0 );
	}
	fc = fb;
	
	for ( iter=1; iter <= ITMAX; iter++ ) 
	{
		 
		if ( ( fb > 0.0 && fc > 0.0 ) || ( fb < 0.0 && fc < 0.0 ) ) 
		{
			c  = a;
			fc = fa;
			e  = d = b - a;
		}
		if (fabs(fc) < fabs(fb)) {
			a  = b;
			b  = c;
			c  = a;
			fa = fb;
			fb = fc;
			fc = fa;
		}
		tol1 = 2.0*EPS*fabs(b)+0.5*xtol;
		xm   = 0.5*(c-b);
 	
		if (fabs(xm) <= tol1 || fb == 0 )   
		{
 			return b; 
		}
		if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) 
		{
			s = fb/fa;
			if (a == c) 
			{
				p = 2.0*xm*s;
				q = 1.0 - s;
			} else {
				q = fa/fc;
				r = fb/fc;
				p = s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
				q = (q-1.0)*(r-1.0)*(s-1.0);
			}
			if (p > 0.0) 
				q = -q;
			p    = fabs(p);
			min1 = 3.0*xm*q-fabs(tol1*q);
			min2 = fabs(e*q);
			if (2.0*p < (min1 < min2 ? min1 : min2))
			{
				e = d;
				d = p/q;
			}
			else 
			{
				d = xm;
				e = d;
			}
		} 
		else 
		{
			d = xm;
			e = d;
		}
		a  = b;
		fa = fb;
		if ( fabs(d) > tol1 )
			b += d;
		else
			b += SIGN(tol1,xm);
		fb = func( b, u, phi );
	}
    printf("Maximum number of iterations exceeded in zbrent");
	 
	return 0.0;
}
#undef ITMAX
#undef EPS
#undef NRANSI
 
 
/*============================================================================
 *
 * Name:: f
 *
 * Description::   calculate LHS of the equality of function for T2 and T3
 *      
 *
 *============================================================================*/

double f(double t)
{  
	return(gamma1*t + gamma2*pow(t,gamma3));
}


/*============================================================================
 *
 * Name:: func
 *
 * Description::   calculate the equality of function for T2 and T3
 *      
 *
 *============================================================================*/

double func( double t, double u, double phi )
{
	double f(double);
    double rhs;

    rhs = -log(u)/phi;
	return ( f( t ) - rhs );
} 
 
/*============================================================================
 *
 * Name:: gen_animal_sort
 *
 * Description::   to preassign randomly sacrifice animals 
 *      
 *
 * Inputs::
 *       index    -  index of the animal
 *       N        -  number of animals
 *       temp     -  temporary array ( see petotest )
 *       A        -  temporary array ( see petotest )
 *
 *
 * Return Value::
 *      index
 *
 *
 *============================================================================*/

void gen_animal_sort(int index[], double temp[], double A[], int N)
{
 
	int l, m;
     
    for(l = 0; l < N; l++)
	{
        temp[l] = getUniform();
        A[l] = temp[l];
    }

    qsort(A, N - 1, sizeof(double), CompareDbl); 

    for(l = 0; l < N; l++)
        for(m = 0; m < N; m++)
            if(temp[l] == A[m])
			{
                index[l] = m+1;
                m = N;
             }
}

/*============================================================================
 *
 * Name:: quicksort
 *
 *
 * Note::
 *      Quicksort was created by C. Anthony R. Hoare and described in his 
 *      1962 paper "Quicksort" (Computer Journal, Vol. 5, No. 1) 
 *      
 *      Kelly, A. and Pohl, I., A Book on C: programming in C - 2nd Ed. (1990). The
 *      Benjamin/Cummings Publishing Company, Inc.
 *============================================================================*/

void quicksort(double *left, double *right)
{
	double *p, pivot;

    if(find_pivot(left, right, &pivot) == yes)
	{
		p = partition(left, right, pivot);
        quicksort(left, p-1);
        quicksort(p, right);
    }
}

/*============================================================================
 *
 * Name:: yes_no find_pivot
 *
 * Description::   This function chooses, if possible, one of the elements of 
 *				   the array to be a "pivot element."
 *      
 *
 * Note::
 *      Quicksort was created by C. Anthony R. Hoare and described in his 
 *      1962 paper "Quicksort" (Computer Journal, Vol. 5, No. 1) 
 *
 *      Kelly, A. and Pohl, I., A Book on C: programming in C - 2nd Ed. (1990). The
 *      Benjamin/Cummings Publishing Company, Inc.
 *============================================================================*/

static yes_no find_pivot(double *left, double *right, double *pivot_ptr)
{
	double a, b, c, *p;

        a = *left;      /* left value */
        b = *(left + (right - left)/2); /*middle value */
        c = *right;     /* right value */
        order3(a, b, c);        /* order these three values */
        if(a < b){ /* pivot will be higher of 2 values */
                *pivot_ptr = b;
                return(yes);
        }
        if(b < c){
                *pivot_ptr = c;
                return(yes);
        }
        for(p = left+1; p <= right; ++p)
                if(*p != *left){
                        *pivot_ptr = (*p < *left) ? *left: *p;
                        return(yes);
                }
        return(no);     /* all elements have the same value */
}

/*============================================================================
 *
 * Name:: partition
 *
 * Description::   This function is used to rearrange the array so that the 
 *                 first part consists of elements of whose values are less than 
 *                 the pivot and the remaining part consists of elements of whose 
 *                 values are greater than or equal to the pivot.  In addition, 
 *                 this function returns a pointer to an element in the array.
 *                 So, elements to the left of the pointer all have value less 
 *                 than the pivot, and elements to the right of the pointer, as 
 *                 well as the element pointed to, all have value greater than or 
 *                 equal to the pivot
 *      
 *
 * Note::
 *      Quicksort was created by C. Anthony R. Hoare and described in his 
 *      1962 paper "Quicksort" (Computer Journal, Vol. 5, No. 1) 
 *
 *      Kelly, A. and Pohl, I., A Book on C: programming in C - 2nd Ed. (1990). The
 *      Benjamin/Cummings Publishing Company, Inc.
 *============================================================================*/

double *partition(double *left, double *right, double pivot)
{
 
        while(left <= right){
                while(*left < pivot)    ++left;
                while(*right >= pivot)  --right;
                if(left < right){
                        swap(*left, *right);
                        ++left; --right;
                }
        }
        return(left);
}

/*============================================================================
 *
 * Name:: CompareDbl
 *
 * Description::
 *     Used to compare two double values and determine whether the first
 *     element in the comparison is considered less than, equal to, or
 *     greater than the second element (like strcmp()).
 *
 * Inputs::
 *     pv1 - Pointer to first double value.
 *     pv2 - Pointer to second double value.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns 1 if second element is larger, -1 if second element is smaller,
 *     or 0 in the unlikely event that both elements are the same.
 *
 *============================================================================*/

static int CompareDbl (const void *pv1, const void *pv2)
{
    const double *ps1 = (double *)pv1;
    const double *ps2 = (double *)pv2;

    assert (ps1 != NULL);
    assert (ps2 != NULL);

    if (*ps1 > *ps2)
        return 1;
    else if (*ps1 < *ps2)
        return -1;
    else
        return 0;
}

/*============================================================================
 *
 * Name:: CompareInt
 *
 * Description::
 *     Used to compare two integer values and determine whether the first
 *     element in the comparison is considered less than, equal to, or
 *     greater than the second element (like strcmp()).
 *
 * Inputs::
 *     pv1 - Pointer to first int value.
 *     pv2 - Pointer to second int value.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns 1 if second element is larger, -1 if second element is smaller,
 *     or 0 in the unlikely event that both elements are the same.
 *
 *============================================================================*/

static int CompareInt (const void *pv1, const void *pv2)
{
    const int *ps1 = (int *)pv1;
    const int *ps2 = (int *)pv2;

    assert (ps1 != NULL);
    assert (ps2 != NULL);

    if (*ps1 > *ps2)
        return 1;
    else if (*ps1 < *ps2)
        return -1;
    else
        return 0;
}

 
/*============================================================================
 *
 * Name:: GetAverages
 *
 * Description::
 *     Used to  compute and display the averages
 *
 * Inputs::
 *     averageInput - struct of type averages_t
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void GetAverages(const averages_t averageInput )
{
    register int i;
	register int k;
	int numGroups; 
	int nit; 
	int intvl_flg;

	numGroups = averageInput.numGroups;
	nit       = averageInput.nit;
	intvl_flg = averageInput.intvl_flg;

    for(i = 0; i < numGroups; i++)
	{
	    printf("dose group %d:\n", i);
        
        printf("average tumor rate = %2.4f\n", averageInput.avg_t1[i]/nit);
        printf("average competing risks survival rate = %2.4f\n",
			1.0 - averageInput.avg_t3[i]/nit);
	    printf("average lethality = %2.4f\n", 
			averageInput.avg_leth[i]/averageInput.divisor[i]);
		printf("\n");

	    fprintf(outfile,"dose group %d:\n", i);
        
        fprintf(outfile,"average tumor rate = %2.4f\n", averageInput.avg_t1[i]/nit);
        fprintf(outfile,"average competing risks survival rate = %2.4f\n",
			 1.0 - averageInput.avg_t3[i]/nit);
	    fprintf(outfile,"average lethality = %2.4f\n", 
			averageInput.avg_leth[i]/averageInput.divisor[i]);
		fprintf(outfile,"\n");
        
		/* To print avg. dose rates */
	    fprintf(outfile,"sacrifice time\t d\t a1\t b1\t a2\t b2\n");
		printf("sacrifice time  d       a1      b1      a2      b2\n");
	    
		for( k=0; k < intvl_flg; k++)
		{
	        fprintf(outfile,"%-14d\t %4.4f\t %4.4f\t %4.4f\t %4.4f\t %4.4f\n",
			 averageInput.sactime[k],	
			(averageInput.d_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.a1_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.b1_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.a2_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.b2_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]));
             
			printf("%-14d  %4.4f  %4.4f  %4.4f  %4.4f  %4.4f\n", 
            averageInput.sactime[k],
			(averageInput.d_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.a1_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.b1_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.a2_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]),
			(averageInput.b2_avg[i][k]/((double) (nit)))/((double) averageInput.N[i]));
			
	    }
	    fprintf(outfile,"\n");
		printf("\n");
	    /* updated on 9/26/02 */   
	} 

}

/*============================================================================
 *
 * Name:: CalculatePower
 *
 * Description::
 *     Used to compute and display the power value
 *
 * Inputs::
 *     (None)
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void CalculatePower()
{

	if ( nit_comb > 0 )
	{
		 
		printf("Positive Trend (Power):\t%2.4f\n",pwr_trend_comb/nit_comb);
	
		fprintf(outfile,
			"Positive Trend (Power):\t%2.4f\n",pwr_trend_comb/nit_comb);
	}
	else
	{
			printf("Positive Trend (Power) could not be calculated due to insufficient number of simulations!");
			fprintf(outfile,"Positive Trend (Power) could not be calculated due to insufficient number of simulations!");
	}
	
	
}

/*============================================================================
 *
 * Name:: InitialProcess
 *
 * Description::
 *     Used to extract the user input data from the xmlInputDataStruct,
 *     perform some input validation, allocate memory for the dynamic
 *     array elemnts of gendataInputStruct, petotestInputStruct and averages   
 *     data structs and intialize the struct elements  
 *
 * Inputs::
 *     data         - data struct of type xmlInputDataStruct
 *	   genInput     - pointer to gendataInputStruct data struct
 *     petoInput    - pointer to petotestInputStruct data struct
 *     averageInput - pointer to averages data struct
 *
 * Outputs::
 *     genInput      
 *     petoInput     
 *     averageInput  
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void InitialProcess ( xmlInputDataStruct data, 
					 gendataInputStruct* genInput, 
					 petotestInputStruct * petoInput,
					 averages_t * averageInput)
{
	register int i;
	register int j;
	register int k;
    int nit;
    int selectseed;
	int NTP;
	int intvl_flg;
	int fix_intvl;
	int rn;
	int N_tot = 0;
    int test_side;
	int numGroups;
	int dose_group;
	int maxintvl = 0;
    double test_level;
	double pt1;
	double delta2;
	double phi2;
	int* animal;
	double * hazardrt;
	double * ed; /* represents e^d */
	double * comp; /* CRSR along with dose groups */
	double * dose_level;
	double *tpt1; /* holds Tumor Onset Pr. of 4 dose groups */
	int * N;
    int ** no_sac_int;
	int * temp;
	int * sactime;
 
	
    /* extract input info from data and initialize struct elemnts */
	numGroups             = data.numberDoseGroups;
	dose_group            = numGroups;
	genInput->dose_group  = numGroups;
	petoInput->numGroups  = numGroups;

	animal            = allocate_integer_vector(0, numGroups);
	petoInput->animal = allocate_integer_vector(0, numGroups);

	maxintvl = 0;
	for( i = 0; i < numGroups; i++)
	{
	    animal[i] = data.numberAnimalsInGroup[i];
	    maxintvl += animal[i];
	}
	 
	nit                   = data.numberDataSets;
	NTP                   = data.numberSacrifices; 
	genInput->NTP         = NTP;
	test_level            = data.testLevel;
    test_side             = data.testSides;
	petoInput->test_side  = data.testSides;
	petoInput->test_level = data.testLevel;
	petoInput->maxintvl   = maxintvl;
	genInput->maxintvl    = maxintvl;
	 
	/* Seed Selection */
	selectseed = data.selectedSeed; 
	
    setSeed( (unsigned long ) selectseed );
 

	if(selectseed < 0)
	{
         throw "Please enter a positive integer value!\n";
    }
	
	printf("\n*** Input Parameters ***\n");
	printf("\n"); 
	 
	fprintf(outfile,"*** Input Parameters ***\n");
	fprintf(outfile,"\n");

	printf("Selected Seed = %d\n", selectseed);
	fprintf(outfile,"Selected Seed = %d\n", selectseed);


	if(numGroups > 1){
	    printf("Number of Groups = %d\n",numGroups);
	    fprintf(outfile,"Number of Groups = %d\n",numGroups);
	}
	else
	{
	     throw "Number of groups should be more than two groups.\n"; 
	}

	/* Allocate vectors of length dose_group */
	dose_level = allocate_real_vector(0,dose_group);
	ed         = allocate_real_vector(0,dose_group);
	comp       = allocate_real_vector(0,dose_group);
	tpt1       = allocate_real_vector(0,dose_group);
 	hazardrt   = allocate_real_vector(0,dose_group);
	N          = allocate_integer_vector(0, numGroups);

	genInput->dose_level = allocate_real_vector(0,dose_group);
	genInput->ed         = allocate_real_vector(0,dose_group);
	genInput->comp       = allocate_real_vector(0,dose_group);
	genInput->tpt1       = allocate_real_vector(0,dose_group);
 	genInput->N          = allocate_integer_vector(0, numGroups);
	 
    /* print dose matric*/
	printf("Dose metric of each group:\n");
	
	for( i=0; i < dose_group; i++ )
	{
	    dose_level[i]           = data.doseLevel[i]; 
		genInput->dose_level[i] = data.doseLevel[i]; 

	    fprintf(outfile,"Dose metric of group %d = %4.2f\n",i,dose_level[i]);
	    printf("%4.2f\t",dose_level[i]);
	}
	 
	/* print number of animals */
	printf("\nNumber of animals in each group\n");
	
	N_tot = 0;
	 
	for(i = 0; i < dose_group; i++)
	{
	    N[i]            = data.numberAnimalsInGroup[i];
		genInput->N[i]  = data.numberAnimalsInGroup[i];
	    printf("%d\t",N[i]);
	    fprintf(outfile,"Number of animals in group %d = %d\n",i,N[i]);
	    if(N[i] < 30)
		{
	         throw "Number of animals in group  is too small.Must be >= 30!\n";
		}
	    N_tot += N[i];	    
	}
	genInput->N_tot = N_tot;
	
    printf("\n");
	fprintf(outfile,"\n");

	animals = (animalinfo *) malloc((unsigned) N_tot*
				      sizeof(animalinfo));

	for ( i = 0; i < N_tot; i++ )
	{
		animals[i].d_intvl_time = 0;
		animals[i].die          = 0;
		animals[i].sac          = 0;
		animals[i].t1           = 0;
		animals[i].t3           = 0;
		animals[i].tt           = 0;
	}
	/* number of sacrifices including a terminal sacrifice */
	NTP            = data.numberSacrifices; 
	genInput->NTP = data.numberSacrifices;

	/* print number of sacrifices */
	printf("Number of sacrifices including a terminal sacrifice = %d\n",NTP);
	fprintf(outfile,"Number of sacrifices including a terminal sacrifice = %d\n",NTP);
	if(NTP < 0)
	{
         throw "Invalid number of sacrifices" ;
	}
	else if (NTP == 1)
	{
		intvl_flg            = 4;
		fix_intvl            = 4;
		genInput->fix_intvl  = 4;
		genInput->intvl_flg  = 4;
		intvl_inc            = 4;
		no_sac_int = mem_alloc_mat_int(0,dose_group,0,4);
		genInput->no_sac_int = mem_alloc_mat_int(0,dose_group,0,4);
		sactime              = allocate_integer_vector(0,4);
		genInput->sactime    = allocate_integer_vector(0,4);
	}
	else if (NTP > 1)
    {
		intvl_flg            = NTP;
		fix_intvl            = NTP;
		genInput->fix_intvl  = NTP;
		genInput->intvl_flg  = NTP;
        intvl_inc            = NTP;
		/* Allocate vectors of length NTP */
		no_sac_int = mem_alloc_mat_int(0,dose_group,0,NTP);
		genInput->no_sac_int = mem_alloc_mat_int(0,dose_group,0,NTP);
		temp                 = allocate_integer_vector(0,NTP);
		sactime              = allocate_integer_vector(0,NTP);
		genInput->sactime    = allocate_integer_vector(0,NTP);
	}
	else
	{
         throw "Error in input data!";
	}
	 
	
    /* print interim sacrifice times */
	printf("Sacrifice time points in weeks:\n");
	fprintf(outfile,"Sacrifice time points in weeks:\n"); 
	if(NTP == 1)
	{
		sactime[3] = data.durationOfStudy;
		sactime[0] = sactime[3]/2;
		sactime[1] = sactime[0] + sactime[3]/4;
		sactime[2] = sactime[1] + sactime[3]/8;
        rn         = sactime[3];

		genInput->sactime[3] = data.durationOfStudy;
		genInput->sactime[0] = genInput->sactime[3]/2;
		genInput->sactime[1] = genInput->sactime[0] + genInput->sactime[3]/4;
		genInput->sactime[2] = genInput->sactime[1] + genInput->sactime[3]/8;
        genInput->rn         = sactime[3];
	}
	else if (NTP > 1)
	{
		for(i=0;i<(NTP-1);i++)
		{  
			sactime[i] = data.sacrificeTimepoint[i];
			genInput->sactime[i] = data.sacrificeTimepoint[i];
		}
		rn           = data.durationOfStudy;	
		genInput->rn = data.durationOfStudy;

		sactime[NTP-1] = rn;
		genInput->sactime[genInput->NTP - 1] = genInput->rn;

		for(i=0;i<NTP;i++)
		{
			printf("%d\t",sactime[i]);
			fprintf(outfile,"Sacrifice time point %d = %d\n",i,sactime[i]);
		}
	}
	else
	{
         throw "Error in input data!";
	}

	/* print studyduration */
	printf("\nStudy duration = %d weeks\n", rn);
	fprintf(outfile,"\nStudy duration = %d weeks\n", rn);
	
	/* print interim sacrifices */
    printf("Number of INTERIM sacrificed animals in each interval:\n");
	fprintf(outfile,"Number of INTERIM sacrificed animals in each interval:\n");
	 
	for (i=0; i < dose_group; i++)
	{
		if(NTP > 1)
		{
		    for(j=0; j<NTP-1; j++)
			{
				temp[j] = data.numberInterimSacrifices[i][j];
				printf("%d\t",temp[j]);
				fprintf(outfile,"%d\t",temp[j]);
				no_sac_int[i][j] = 0;
				genInput->no_sac_int[i][j] = 0;
				for( k=0; k<=j; k++)
				{
					no_sac_int[i][j]            += temp[k];
					genInput->no_sac_int[i][j]  += temp[k];
				}
				if(no_sac_int[i][j] > N[i])
				{
				    throw("Total number of sacrified animals should not exceed the number of animals in each group!\n");
                }
			}

			no_sac_int[i][NTP-1]                         = N[i];
			genInput->no_sac_int[i][genInput->NTP - 1] = genInput->N[i];
			printf("\n");
			fprintf(outfile,"\n");
		}
		else if (NTP ==1)
		{
			no_sac_int[i][0] = 0;
			no_sac_int[i][1] = 0;
			no_sac_int[i][2] = 0;
			no_sac_int[i][3] = N[i];

			genInput->no_sac_int[i][0] = 0;
			genInput->no_sac_int[i][1] = 0;
			genInput->no_sac_int[i][2] = 0;
			genInput->no_sac_int[i][3] = N[i];
		}
		else
		{
             throw "Error in input data!";
		}
	
	}
 
	/* background tumor onset probability */
	pt1 = data.tumorOnsetProbability; 
	genInput->pt1 = data.tumorOnsetProbability; 
	printf(
		"Background tumor onset probability at the end of the study = %4.2f\n",pt1);
	fprintf(outfile,
		"Background tumor onset probability at the end of the study = %4.2f\n",pt1);
	if(pt1 > 1.0)
	{
        throw "Probability should not exceed 1.0.\n";
	}
	 
	/* Tumor onset distribution */
	delta2            = data.shapeParameter; 
	genInput->delta2 = data.shapeParameter; 
	printf(
		"Tumor onset distribution assumed: Weibull with a shape parameter %4.2f\n", delta2);
	fprintf(outfile,
		"Tumor onset distribution assumed: Weibull with a shape parameter %4.2f\n", delta2);
    
	/* print hazard ratios */
    printf("Hazard ratio(s) of dose vs. control group\n"); 
	fprintf(outfile,"Hazard ratio(s)\n"); 
	 	
	for(i=0;i<dose_group-1;i++)
	{
	    hazardrt[i]            = data.hazardRate[i];
		printf("%4.2f\t", hazardrt[i]);
	    fprintf(outfile,"dose %d vs. control = %4.2f\n", i+1, hazardrt[i]);
	}
	 
    /* print CRSR */
	printf("\nCompeting Risks Survival Rate (CRSR) for each group:\n");
	fprintf(outfile,"\nCompeting Risks Survival Rate (CRSR) for each group:\n");
	 
	for(i=0;i<dose_group;i++)
	{
	    comp[i]            = data.competingRiskSurvRate[i];
		genInput->comp[i]  = data.competingRiskSurvRate[i];

	    printf("%4.2f\t",comp[i]);
	    fprintf(outfile,"group %d = %4.2f\n",i,comp[i]);
	    if(comp[i] > 1.0)
		{
            throw "\nProbability should not exceed 1.0. \n";
	    }

	}
	 
	/* assign Q(t_max)_control as f_gam */
	f_gam = comp[0];

	/* tumor lethality */
	phi2            = data.tumorLethalityParameter;
	genInput->phi2 = data.tumorLethalityParameter;

	printf("\nTumor lethality parameter entered = %4.2f\n", phi2);
	fprintf(outfile,"\nTumor lethality parameter entered = %4.2f\n", phi2);
    
	
	/* level of the test */
	test_level = data.testLevel; 
	
	printf("Level of the test = %4.2f\n",test_level);
	fprintf(outfile,"Level of the test = %4.2f\n",test_level);
   
	/* one-sided test or two-sided test */
	 
	
	printf("One-sided or two-sided test = %d sided test\n", test_side);
	fprintf(outfile,"One-sided or two-sided test = %d sided test\n", test_side);

	/* number of data set */
	nit = data.numberDataSets; 
	printf("Number of simulation runs = %d\n",nit);
	fprintf(outfile,"Number of simulation runs = %d\n",nit);
	printf("\n");
	if(nit > 10000)
	{
        throw "The number of data sets is unnecessarily large.\n";  
	}
	 
	fprintf(outfile,"\n*** Simulation Results ***\n");
    printf("\n*** Simulation Results ***\n");
    printf("\n");
	fprintf(outfile,"\n");

	// initialize genInput elements

 
	if(f_gam >= exp(-gamma1*rn))
	{
         throw "Competing Risks Survival Rate in control group is too big!\n";
	}
	gamma3 = log((-log(f_gam) - gamma1*rn)/gamma2)/log(rn);
	
	genInput->ed[0]=1.0;
	for( i=0; i < numGroups-1; i++)
	{
		genInput->ed[i+1] = hazardrt[i];
	}

	/* tpt1[] holds the tumor onset prob. on each group */
	for(i = 0; i < numGroups; i++)
	{
		genInput->tpt1[i] = 1.0-exp(-(genInput->ed[i])*delta1);
	}


	//Alocating globals
	
    no_w_it    = mem_alloc_mat_int(0,numGroups-1,0,intvl_inc-1);
	no_dead_it = mem_alloc_mat_int(0,numGroups-1,0,intvl_inc-1);
	no_w_t     = mem_alloc_mat_int(0,numGroups-1,0,maxintvl-1);
	no_surv    = mem_alloc_mat_int(0,numGroups-1,0,maxintvl-1);
	y_g_sum    = allocate_integer_vector(0, intvl_inc-1);
	n_g_sum    = allocate_integer_vector(0, intvl_inc-1);
	y_g_sum_ft = allocate_integer_vector(0,maxintvl-1);
	n_g_sum_ft = allocate_integer_vector(0,maxintvl-1);
	y_poly3    = allocate_integer_vector(0,numGroups-1);
	dose       = allocate_real_vector(0,numGroups-1);
    D          = allocate_real_vector(0,numGroups-1);
	D_ft	   = allocate_real_vector(0,numGroups-1);
	D_tot      = allocate_real_vector(0,numGroups-1);
	K          = memory_alloc_mat(0,numGroups-1,0,intvl_inc-1);
	K_ft       = memory_alloc_mat(0,numGroups-1,0,maxintvl-1);
	E          = memory_alloc_mat(0,numGroups-1,0,intvl_inc-1);
	E_ft       = memory_alloc_mat(0,numGroups-1,0,maxintvl-1);
	V_tot      = memory_alloc_mat(0,numGroups-1,0,numGroups-1);
	V_a        = memory_alloc_mat(0,numGroups-1,0,numGroups-1);
	V_b        = memory_alloc_mat(0,numGroups-1,0,numGroups-1);

	// initialize global variables
	pwr_trend_comb = 0.0; 
	pwr_poly3      = 0.0;
    pwr_ca         = 0.0; 
	pwr_trend_ft   = 0.0; 
	pwr_trend_inc  = 0.0;
	nit_ft         = 0.0; 
	nit_inc        = 0.0; 
	nit_comb       = 0.0;
     
	for ( i = 0; i < numGroups; i++)
	{
		dose[i] = data.doseLevel[i];
	}

	// allocate memory and initialize the average arrays
	InitializeAverages( averageInput, sactime, intvl_flg, numGroups, NTP, nit);


	// free local arrays  
	free_integer_vector(animal, 0);
	
	free_real_vector(dose_level,0);
	free_real_vector(ed,0);
	free_real_vector(comp,0);
	free_real_vector(tpt1,0);
	free_real_vector(hazardrt, 0); 
	free_integer_vector(N, 0);
	


}

/*============================================================================
 *
 * Name:: InitializeAverages
 *
 * Description::
 *      Allocates memory for the dynamic array elements of averages struct 
 *      and initializes them
 *
 * Inputs::
 *     averageInput - pointer to averages data struct
 *     st           - pinter to array of integers, containing the sacrifice times 
 *	   intvl_flg    - integer 
 *     numGroups    - integer - number of the dose groups 
 *     NTP          - integer - number of sacrifices
 *     nit          - integer - number of simulations
 *
 * Outputs::
 *     averageInput 
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void InitializeAverages( averages_t* averageInput, 
						 const int* st,
						 const int intvl_flg,
						 const int numGroups,
						 const int NTP, 
						 const int nit)
{
	register int i;
	register int j;
	
	averageInput->numGroups = numGroups;
	averageInput->nit       = nit;
	averageInput->intvl_flg = intvl_flg;
	averageInput->NTP       = NTP;
  
    averageInput->avg_leth = allocate_real_vector(0, numGroups);
	averageInput->avg_t1   = allocate_real_vector(0, numGroups);
	averageInput->avg_t3   = allocate_real_vector(0, numGroups);
	averageInput->divisor  = allocate_real_vector(0, numGroups);
    averageInput->N        = allocate_integer_vector(0, numGroups);
		
	averageInput->d_avg  = mem_alloc_mat_int(0, numGroups, 0, intvl_flg);
	averageInput->a1_avg = mem_alloc_mat_int(0, numGroups, 0, intvl_flg);
	averageInput->b1_avg = mem_alloc_mat_int(0, numGroups, 0, intvl_flg);
	averageInput->a2_avg = mem_alloc_mat_int(0, numGroups, 0, intvl_flg);
	averageInput->b2_avg = mem_alloc_mat_int(0, numGroups, 0, intvl_flg);


	for( i = 0; i< numGroups; i++)
	{
	    for( j = 0; j< intvl_flg; j++)
		{
	        averageInput->d_avg[i][j]  = 0; 
			averageInput->a1_avg[i][j] = 0; 
			averageInput->b1_avg[i][j] = 0;
		    averageInput->a2_avg[i][j] = 0; 
			averageInput->b2_avg[i][j] = 0;
	    }
	}

   
	for(i = 0; i < numGroups; i++)
	{
		 averageInput->avg_leth[i] = 0.0; 
		 averageInput->avg_t1[i]   = 0.0; 
		 averageInput->avg_t3[i]   = 0.0;
         averageInput->divisor[i]  = 0.0;
     }


		// allocate memory for sactime
    if (NTP == 1)
	{
		 
		averageInput->sactime  = allocate_integer_vector(0,4);

		for ( i = 0; i < 4; i++)
		{
			averageInput->sactime[i] = st[i];
		}
		 
	}
	else if (NTP > 1)
    {
	 
		averageInput->sactime  = allocate_integer_vector(0,NTP);

		for ( i = 0; i < NTP; i++)
		{
			averageInput->sactime[i] = st[i];
		}
	 
	}
	return;
}

/*============================================================================
 *
 * Name:: AllocateDataStructs
 *
 * Description::
 *      Allocates memory for the dynamic array elements of sampData and sampftData
 *      data structs
 *
 * Inputs::
 *     sampData   - pointer to sampleData_t data struct
 *	   numGroups  - number of dose groups 	
 *	   maxintvl   - integer - maximum number of intervals ( also max number of
 *                      animals)
 * 
 * Outputs::
 *     sampData 
 *     sampftData
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void AllocateDataStructs( sampleData_t** sampData, 
						  int numGroups,
				          int maxintvl)
{
     register int jj;

	 for ( jj = 0; jj < numGroups; jj++)
	{
	    // Max amount of memory allocated!
        (*sampData)[jj].numberDead                = allocate_integer_vector(1, maxintvl);
		(*sampData)[jj].numDeadWithTumor          = allocate_integer_vector(1, maxintvl);
		(*sampData)[jj].numDeadWithoutTumor       = allocate_integer_vector(1, maxintvl);
		(*sampData)[jj].numSacrificesWithTumor    = allocate_integer_vector(1, maxintvl);
		(*sampData)[jj].numSacrificesWithoutTumor = allocate_integer_vector(1, maxintvl);
		(*sampData)[jj].intervalEndpoints         = allocate_real_vector(1, maxintvl); 
	}

}

/*============================================================================
 *
 * Name:: DeallocateDataStructs
 *
 * Description::
 *      Frees the memory allocated for the dynamic array elements of sampData
 *      and sampftData data structs
 *
 * Inputs::
 *     sampData   - sampleData_t data struct
 *     numGroups  - number of dose groups
 * 
 * Outputs::
 *     (None) 
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void DeallocateDataStructs( sampleData_t* sampData, int numGroups )
{
	register int jj;
    for(jj = 0; jj < numGroups ; jj++)
	{
		free_integer_vector (sampData[jj].numberDead, 1);
		free_integer_vector (sampData[jj].numDeadWithTumor, 1);
		free_integer_vector (sampData[jj].numDeadWithoutTumor, 1);
		free_integer_vector (sampData[jj].numSacrificesWithTumor, 1);
		free_integer_vector (sampData[jj].numSacrificesWithoutTumor, 1);
		free_real_vector    (sampData[jj].intervalEndpoints, 1); 

	}
}

/*============================================================================
 *
 * Name:: FinalProcess
 *
 * Description::
 *      Frees the memory allocated for the dynamic array elements of genInput,
 *      petoInput and averageInput data structs as well as the memory allocated
 *      to global arrays
 *
 * Inputs::
 *     numGroups     - integer - number of dose groups 
 *     genInput      - gendataInputStruct_t data struct
 *     petoInput     - petotestInputStruct_t data struct
 *     averageInput  - averages_t data struct
 * 
 * Outputs::
 *     (None) 
 *
 * Return Value::
 *     (None) 
 *
 *============================================================================*/

void FinalProcess( int numGroups, 
				   gendataInputStruct_t genInput, 
				   petotestInputStruct_t petoInput, 
				   averages_t averageInput )
{
    // free globals
	free(animals);
	free_int_matrix (no_w_it, 0, numGroups-1, 0);
	free_int_matrix (no_dead_it, 0, numGroups-1, 0);
	free_int_matrix (no_w_t, 0, numGroups-1, 0);
	free_int_matrix (no_surv, 0, numGroups-1, 0);
	free_integer_vector (y_g_sum, 0);
	free_integer_vector (n_g_sum, 0);
	free_integer_vector (y_g_sum_ft, 0);
	free_integer_vector (n_g_sum_ft, 0);
	free_integer_vector (y_poly3, 0);
	free_real_vector (dose, 0);
	free_real_vector (D, 0);
	free_real_vector (D_ft, 0);
	free_real_vector (D_tot, 0);
	free_real_matrix (K, 0, numGroups-1, 0);
	free_real_matrix (K_ft, 0, numGroups-1,0);
	free_real_matrix (E,0, numGroups-1, 0);
	free_real_matrix (E_ft, 0, numGroups-1, 0);
	free_real_matrix (V_tot, 0, numGroups-1, 0);
	free_real_matrix (V_a, 0, numGroups-1, 0);
	free_real_matrix (V_b, 0, numGroups-1, 0);

	// free petoInput.animals
	free_integer_vector (petoInput.animal, 0);

	// free averageInput pointers
    free_real_vector(averageInput.avg_leth, 0);
	free_real_vector(averageInput.avg_t1, 0);
	free_real_vector(averageInput.avg_t3, 0);
	free_real_vector(averageInput.divisor, 0);
	free_integer_vector(averageInput.N, 0);

	free_integer_vector(averageInput.sactime,0); 
	free_int_matrix(averageInput.d_avg, 0, numGroups, 0);
	free_int_matrix(averageInput.a1_avg, 0, numGroups, 0);
	free_int_matrix(averageInput.b1_avg, 0, numGroups, 0);
	free_int_matrix(averageInput.a2_avg, 0, numGroups, 0);
	free_int_matrix(averageInput.b2_avg, 0, numGroups, 0);

	// free genInput pointers
	free_real_vector(genInput.dose_level, 0);
	free_real_vector(genInput.ed, 0);
	free_real_vector(genInput.comp, 0);
	free_real_vector(genInput.tpt1, 0);
    free_integer_vector(genInput.N, 0);
	free_integer_vector(genInput.sactime, 0);
	free_int_matrix (genInput.no_sac_int, 0, numGroups-1, 0);
}

 
