
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "driveP.h"


/*============================================================================
 *
 * Name:: memory_alloc_mat
 *
 * Description:: allocate a real matrix of range [lr..ur][lc..uc] 
 *      
 *
 * Inputs::
 *      lr      -  starting row index
 *      ur      -  ending row index 
 *      lc      -  starting column index
 *      uc      -  ending column index
 *
 *
 * Return Value::
 *      a pointer to a 2dimentional array of double
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

double **memory_alloc_mat(int lr, int ur, int lc, int uc)
{
	int i;
	double **pt;

	pt = (double **)malloc((unsigned) (ur-lr+1)*sizeof(double*));
	if(!pt){	printf("failure in memory_alloc_mat1()\n");
			return NULL;
	}
	pt -= lr;

	for(i = lr; i <= ur; i++){
		pt[i] = (double *)malloc((unsigned) (uc-lc+1)*sizeof(double));
		if(!pt[i]){	printf("failure in memory_alloc_mat2()\n");
				return NULL;
		}
		pt[i] -= lc;
	}
	return(pt);
}	

/*============================================================================
 *
 * Name:: mem_alloc_mat_int
 *
 * Description:: allocate an integer matrix of range [lr..ur][lc..uc] 
 *      
 *
 * Inputs::
 *      lr      -  starting row index
 *      ur      -  ending row index 
 *      lc      -  starting column index
 *      uc      -  ending column index
 * 
 * Outputs::
 *     (none)
 *
 * Return Value::
 *      a pointer to a 2dimentional array of double
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

int **mem_alloc_mat_int(int lr, int ur, int lc, int uc)
{
 	int i;
	int **pt;

	pt = (int **)malloc((unsigned) (ur-lr+1)*sizeof(int*));
	if(!pt){	printf("failure in memory_alloc_mat1()\n");
			return NULL;
	}
	pt -= lr;

	for(i = lr; i <= ur; i++){
		pt[i] = (int *)malloc((unsigned) (uc-lc+1)*sizeof(int));
		if(!pt[i]){	printf("failure in memory_alloc_mat2()\n");
				return NULL;
		}
		pt[i] -= lc;
	}
	return(pt);
}	



/*============================================================================
 *
 * Name:: free_real_matrix
 *
 * Description:: deallocate a real matrix of range [lr..ur][lc..uc] 
 *      
 *
 * Inputs::
 *      m       -  real matrix 
 *      lr      -  starting row index
 *      ur      -  ending row index 
 *      lc      -  starting column index
 *      uc      -  ending column index
 * 
 * Outputs::
 *     (none)
 *
 * Return Value::
 *     (none) 
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

void free_real_matrix(double **m, int lr, int ur, int lc){
	int i;

	for(i = ur; i >= lr; i--)	free((char*) (m[i]+lc));
	free((char*) (m+lr));
}

/*============================================================================
 *
 * Name:: free_int_matrix
 *
 * Description:: deallocate an integer matrix of range [lr..ur][lc..uc] 
 *      
 *
 * Inputs::
 *      m       -  integer matrix 
 *      lr      -  starting row index
 *      ur      -  ending row index 
 *      lc      -  starting column index
 *      uc      -  ending column index
 * 
 * Outputs::
 *     (none)
 *
 * Return Value::
 *     (none) 
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

void free_int_matrix(int **m, int lr, int ur, int lc){
	int i;

	for(i = ur; i >= lr; i--)	free((char*) (m[i]+lc));
	free((char*) (m+lr));
}

/*============================================================================
 *
 * Name:: allocate_integer_vector
 *
 * Description:: allocate an integer vector of range [l..u]
 *      
 *
 * Inputs::
 *       l            -  beginning of range 
 *       u            -  end of range
 * Outputs::
 *     (none)
 *
 * Return Value::
 *      a pointer to integer
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

int *allocate_integer_vector(int l, int u)
{
	int *pt;

	pt = (int *)calloc((unsigned) (u+1),sizeof(int));
    if(!pt)
	{
        printf("failure in allocate_integer_vector()\n");
       return NULL;
    }
    return(pt);
}

/*============================================================================
 *
 * Name:: allocate_real_vector
 *
 * Description:: allocate a real vector of range [l..u]
 *      
 *
 * Inputs::
 *       l            -  beginning of range 
 *       u            -  end of range
 * Outputs::
 *     (none)
 *
 * Return Value::
 *      a pointer to double
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781. 
 *============================================================================*/

double *allocate_real_vector(int l, int u)
{
	double *pt;
	
	pt = (double *)calloc((unsigned) (u+1),sizeof(double));
    if(!pt)
	{        
		printf("failure in allocate_real_vector()\n");
        return NULL;
    }

    return(pt);
}

/*============================================================================
 *
 * Name:: free_real_vector
 *
 * Description:: frees assigned memory allocations 
 *      
 *
 * Inputs::
 *      v       -  vector to deallocate 
 *      l       -  number of elements of vector v
 *
 * Return Value::
 *     (none) 
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/
 
void free_real_vector(double *v, int l)
{
	free((double*) v); 
}

/*============================================================================
 *
 * Name:: free_integer_vector
 *
 * Description:: frees assigned memory allocations 
 *      
 *
 * Inputs::
 *      v       -  vector to deallocate 
 *      l       -  number of elements of vector v
 *
 * Return Value::
 *     (none) 
 *
 * Reference:
 *      Lau, T. (1995) A Numerical Library in C for Scientists and
 *                     Engineers. CRC Press, Inc. 779-781.
 *============================================================================*/

void free_integer_vector(int *v, int l)
{ 
	free((int*) v ); 
}

 
  